"""
General utilities
"""


# @check_types
# def summarize(df: DataFrame[Snapshot]) -> DataFrame[Summary]:
#     df[Summary.date] = df[Snapshot.timestamp].dt.date
#     return (
#         df.rename(columns={get_name_field_gen_mw(df): Summary.gen_mw})
#         .groupby([Summary.date, Snapshot.fuel], as_index=False)
#         .agg({Summary.gen_mw: "sum"})
#     )


# @check_types
# def collect_summaries(
#     time_units: List[TimeUnit], last_modified: Optional[str] = None
# ) -> DataFrame[Summary]:
#     return concat([time_unit.summary for time_unit in time_units])


# def get_end_of_last_month() -> date:
#     now = datetime.now(TZ)
#     return date(year=now.year, month=now.month, day=1) - timedelta(days=1)



from fastapi import FastAPI
from nysmix.db import DatabasePostgres, File, Snapshot
from sqlmodel import Session, select, DATE, cast, func
from fastapi.middleware.cors import CORSMiddleware

app = FastAPI()

origins = [
    "https://gautsi.static.observableusercontent.com",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


db = DatabasePostgres()


@app.on_event("startup")
def on_startup():
    db.create()


@app.get("/")
async def root():
    return {"message": "Hello World test again"}


@app.get("/files/")
def read_files():
    with Session(db.engine) as session:
        files = session.exec(select(File)).all()
    return files


@app.get("/summary/")
def get_daily_summary():
    with Session(db.engine) as session:
        statement = select(
            cast(Snapshot.timestamp, DATE), Snapshot.fuel, func.sum(Snapshot.gen_mw)
        ).group_by(cast(Snapshot.timestamp, DATE), Snapshot.fuel)
        result = session.exec(statement).all()
    return result



"""
Tools for managing fuel mix monthly zips and daily csvs from NYISO 
"""

from abc import ABC, abstractmethod
from datetime import datetime
from logging import info
from pickle import dumps, loads

from fs_gcsfs import GCSFS

from .settings import SettingsGCS


class Repository(ABC):
    @abstractmethod
    def add(self, name: str, last_modified: datetime, contents: bytes) -> None:
        pass

    @abstractmethod
    def contains(self, name: str) -> bool:
        pass

    @abstractmethod
    def get_last_modifieds(self, name: str) -> set[datetime]:
        pass

    @abstractmethod
    def get(self, name: str, last_modified: datetime) -> bytes:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass


class RepositoryInMemory(Repository):
    def __init__(self):
        self.last_modifieds: dict[str, set[datetime]] = {}
        self.contents: dict[tuple(str, datetime), bytes] = {}

    def add(self, name: str, last_modified: datetime, contents: bytes) -> None:
        if self.contains(name):
            self.last_modifieds[name].add(last_modified)
        else:
            self.last_modifieds[name] = {last_modified}
        self.contents[(name, last_modified)] = contents

    def contains(self, name: str) -> bool:
        return name in self.last_modifieds.keys()

    def get_last_modifieds(self, name: str) -> set[datetime]:
        return self.last_modifieds[name]

    def get(self, name: str, last_modified: datetime) -> bytes:
        return self.contents[(name, last_modified)]

    def clear(self) -> None:
        pass


class RepositoryGCS(Repository):
    def __init__(self, name_bucket: str = SettingsGCS().name_bucket, prefix: str = ""):
        self.fs = GCSFS(name_bucket)
        self.prefix = prefix
        self.name_file_last_modifieds = f"{self.prefix}last_modifieds"
        self.last_modifieds: dict[str, set[datetime]] = (
            loads(self.get_from_name(self.name_file_last_modifieds))
            if self.fs.isfile(self.name_file_last_modifieds)
            else {}
        )

    def add_to_name(self, name: str, contents: bytes) -> None:
        info(f"adding to repo {self}: {name}")
        with self.fs.open(name, "wb") as f:
            f.write(contents)

    def name(self, name: str, last_modified: datetime) -> str:
        return f"{self.prefix}{last_modified.strftime('%Y%m%d_%H%M%S')}_{name}"

    def add(self, name: str, last_modified: datetime, contents: bytes) -> None:
        if name in self.last_modifieds.keys():
            self.last_modifieds[name].add(last_modified)
        else:
            self.last_modifieds[name] = {last_modified}
        self.add_to_name(self.name_file_last_modifieds, dumps(self.last_modifieds))
        self.add_to_name(self.name(name, last_modified), contents)

    def contains(self, name: str) -> bool:
        return name in self.last_modifieds.keys()

    def get_last_modifieds(self, name: str) -> set[datetime]:
        return self.last_modifieds[name]

    def get_from_name(self, name: str) -> bytes:
        info(f"grabbing from repo {self}: {name}")
        with self.fs.open(name, "rb") as f:
            content = f.read()
        return content

    def get(self, name: str, last_modified: datetime) -> bytes:
        return self.get_from_name(self.name(name, last_modified))

    def remove(self, name: str) -> None:
        info(f"removing from repo {self}: {name}")
        self.fs.remove(name)

    def clear(self) -> None:
        for name, dts in self.last_modifieds.items():
            for dt in dts:
                self.remove(self.name(name, dt))
        self.remove(self.name_file_last_modifieds)
        self.last_modifieds = {}



from pydantic import BaseSettings, PostgresDsn


class SettingsGCS(BaseSettings):
    name_bucket: str
    name_project: str
    name_folder_cache: str = "cache"

    class Config:
        env_prefix = "gcs_"


class SettingsPostgres(BaseSettings):
    server: str
    user: str
    password: str
    db: str
    port: str

    @property
    def dsn(self) -> PostgresDsn:
        return PostgresDsn.build(
            scheme="postgresql",
            user=self.user,
            password=self.password,
            host=self.server,
            path=f"/{self.db}",
            port=self.port,
        )

    class Config:
        env_prefix = "postgres_"



"""
general concepts
"""

from enum import Enum, auto
from typing import Dict, Set


class Timezone(Enum):
    EST = "EST"
    EDT = "EDT"


class Fuel(Enum):
    HYDRO = "Hydro"
    WIND = "Wind"
    OTHER_RENEWABLES = "Other Renewables"
    NUCLEAR = "Nuclear"
    NATURAL_GAS = "Natural Gas"
    DUAL_FUEL = "Dual Fuel"
    OTHER_FOSSIL_FUELS = "Other Fossil Fuels"


class SimpleFuel(Enum):
    FOSSIL_FUEL = "Fossil fuel"
    RENEWABLE = "Renewable"
    NUCLEAR = "Nuclear"


MEMBERS: Dict[SimpleFuel, Set[Fuel]] = {
    SimpleFuel.FOSSIL_FUEL: {Fuel.NATURAL_GAS, Fuel.DUAL_FUEL, Fuel.OTHER_FOSSIL_FUELS},
    SimpleFuel.RENEWABLE: {Fuel.HYDRO, Fuel.WIND, Fuel.OTHER_RENEWABLES},
    SimpleFuel.NUCLEAR: {Fuel.NUCLEAR},
}



"""
Tools for working with the fuel mix data in a database
"""

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional

from sqlalchemy import Column, Enum
from sqlalchemy.future import Engine
from sqlmodel import Field, SQLModel, create_engine, Session, select

from nysmix.concepts import Fuel, Timezone

from .settings import SettingsPostgres
from logging import info


class File(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    last_modified: datetime


class Snapshot(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    timestamp: datetime = Field(alias="Time Stamp")
    timezone: Timezone = Field(alias="Time Zone", sa_column=Column(Enum(Timezone)))
    fuel: Fuel = Field(alias="Fuel Category", sa_column=Column(Enum(Fuel)))
    gen_mw: float = Field(alias="Gen MW")

    file_id: Optional[int] = Field(default=None, foreign_key="file.id")

    class Config:
        allow_population_by_field_name = True


class Database(ABC):
    engine: Engine

    def create(self) -> None:
        SQLModel.metadata.create_all(self.engine)

    def add_file(self, file: File) -> int:
        with Session(self.engine) as session:
            info(f"loading file {file}...")
            session.add(file)
            session.commit()
            id = file.id
            assert id is not None
            info(f"...loaded with id {id}.")
            return id

    def has_file(self, file: File) -> bool:
        with Session(self.engine) as session:
            num_files = (
                session.query(File)
                .filter(
                    (File.name == file.name)
                    & (File.last_modified == file.last_modified)
                )
                .count()
            )
        return num_files > 0

    def rm_file(self, file: File) -> None:
        if not self.has_file(file):
            return
        info(f"removing file {file}...")
        with Session(self.engine) as session:
            statement = select(File).where(
                (File.name == file.name) & (File.last_modified == file.last_modified)
            )
            results = session.exec(statement)
            file = results.one()
            session.delete(file)
            session.commit()
            info(f"...removed.")

    def add_snapshots(self, snapshots: list[Snapshot]) -> None:
        with Session(self.engine) as session:
            info("loading snapshots...")
            session.add_all(snapshots)
            session.commit()
            info("...loaded.")

    # @property
    # def max_id(self) -> int:
    #     with Session(self.engine) as session:
    #         max_id = session.query(func.max(Snapshot.id)).scalar()
    #     return max_id or 0


class DatabaseInMemory(Database):
    def __init__(self):
        self.engine = create_engine("sqlite://")


class DatabasePostgres(Database):
    def __init__(self, settings: Optional[SettingsPostgres] = None):
        self.settings = settings if settings else SettingsPostgres()
        self.engine = create_engine(
            self.settings.dsn,
            echo=True,
        )



"""
data implementations that optionally uses joblib caching
"""

from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from io import BytesIO
from logging import info
from typing import List, Set
from zipfile import ZipFile

from pandas import read_csv
from pandera import check_types
from pandera.typing import DataFrame
from sqlmodel import Session

from .base import DayBase, MonthBase
from .concepts import Fuel, Timezone
from .config import DATE_START, TZ, YEAR_START
from .db import Database, DatabaseInMemory, File
from .db import Snapshot as SnapshotSQL
from .repository import Repository, RepositoryInMemory
from .schema import Snapshot


@dataclass
class Day:
    repo: Repository = RepositoryInMemory()
    db: Database = DatabaseInMemory()
    day_base: DayBase = DayBase.from_date(DATE_START)

    @property
    def last_modified(self) -> datetime:
        return max(self.repo.get_last_modifieds(self.month.month_base.name_file_zip))

    @cached_property
    def content(self) -> bytes:
        with BytesIO(
            self.repo.get(self.month.month_base.name_file_zip, self.last_modified)
        ) as f:
            with ZipFile(f) as zf:
                content = zf.read(self.day_base.name_file_csv)
        return content

    @property
    def month(self):  # -> Month
        return Month(month_base=MonthBase.from_date(self.day_base.dt), repo=self.repo)

    def extract(self) -> None:
        info(f"extracting daily csv for {self.day_base.dt.strftime('%Y/%m/%d')}")
        name = self.day_base.name_file_csv
        last_modified = self.last_modified
        if not self.repo.contains(
            name
        ) or last_modified not in self.repo.get_last_modifieds(name):
            self.repo.add(name=name, last_modified=last_modified, contents=self.content)

    @property  # type: ignore
    @check_types
    def df(self) -> DataFrame[Snapshot]:
        return (
            read_csv(BytesIO(self.content), dtype=object)
            .rename(columns={"Gen MWh": "Gen MW"})
            .assign(
                **{
                    Snapshot.timezone: lambda df: df[Snapshot.timezone].map(Timezone),
                    Snapshot.fuel: lambda df: df[Snapshot.fuel].map(Fuel),
                }
            )
        )

    def snapshots(self, file_id: int) -> list[SnapshotSQL]:
        return [
            SnapshotSQL(**info, file_id=file_id)
            for info in self.df.to_dict(orient="records")
        ]

    @property
    def file(self) -> File:
        return File(name=self.day_base.name_file_csv, last_modified=self.last_modified)

    def load(self) -> None:
        info(f"loading snapshots to db {self.db} for day {self.day_base.dt}...")
        id = self.db.add_file(self.file)
        try:
            self.db.add_snapshots(self.snapshots(id))
        except Exception as e:
            info("...could not load snapshots.")
            self.db.rm_file(self.file)
            raise e
        info("...loading snapshots done.")


@dataclass
class Month:
    repo: Repository = RepositoryInMemory()
    db: Database = DatabaseInMemory()
    month_base: MonthBase = MonthBase.from_date(DATE_START)

    @property
    def days(self) -> List[Day]:
        days = []
        current = DayBase.from_date(dt=self.month_base.dt)
        while current.yearmonth == self.month_base:
            if not current.is_before_start and not current.is_future:
                days.append(current)
            current = current.next
        return [Day(day_base=day, repo=self.repo, db=self.db) for day in days]

    def pull(self) -> None:
        info(f"pulling zip for month {self.month_base.year}/{self.month_base.month}")
        name = self.month_base.name_file_zip
        last_modified = self.month_base.last_modified
        if not self.repo.contains(
            name
        ) or last_modified not in self.repo.get_last_modifieds(name):
            self.repo.add(
                name=name, last_modified=last_modified, contents=self.month_base.zip
            )

    def extract(self) -> None:
        info(
            f"extracting daily csvs for month {self.month_base.year}/{self.month_base.month}"
        )
        for day in self.days:
            day.extract()

    def load(self) -> None:
        info(
            f"loading daily csvs into db for month {self.month_base.year}/{self.month_base.month}"
        )
        for day in self.days:
            if not self.db.has_file(day.file):
                day.load()


@dataclass
class Year:
    repo: Repository = RepositoryInMemory()
    db: Database = DatabaseInMemory()
    year: int = YEAR_START

    @property
    def months(self) -> List[Month]:
        return [
            Month(repo=self.repo, db=self.db, month_base=month)
            for month in [
                MonthBase(year=self.year, month=month) for month in range(1, 13)
            ]
            if not month.is_future and not month.is_before_start
        ]

    def pull(self) -> None:
        info(f"pulling monthly zips for year {self.year}")
        for month in self.months:
            month.pull()

    def extract(self) -> None:
        info(f"extracting daily csvs for year {self.year}")
        for month in self.months:
            month.extract()

    def load(self) -> None:
        info(f"loading daily csvs into db for year {self.year}")
        for month in self.months:
            month.load()


@dataclass
class All:
    repo: Repository = RepositoryInMemory()
    db: Database = DatabaseInMemory()

    @property
    def years(self) -> List[Year]:
        now = datetime.now(TZ)
        return [
            Year(repo=self.repo, db=self.db, year=year)
            for year in range(YEAR_START, now.year + 1)
        ]

    def pull(self) -> None:
        info("pulling monthly zips from all years")
        for year in self.years:
            year.pull()

    def extract(self) -> None:
        info("extracting daily csvs from all years")
        for year in self.years:
            year.extract()

    def load(self) -> None:
        info("loading daily csvs into db from all years")
        for year in self.years:
            year.load()



"""general configuration
"""

from datetime import date
from datetime import date, timedelta, timezone


YEAR_START = 2015
MONTH_START = 12
DAY_START = 9
DATE_START = date(year=YEAR_START, month=MONTH_START, day=DAY_START)
FORMAT_DATE = "%Y%m%d"
TZ = timezone(timedelta(hours=-5))



from datetime import datetime
from typing import Annotated

from pandera import Category, Field, SchemaModel
from pandera.typing import Series

from .concepts import Fuel, Timezone


class Snapshot(SchemaModel):
    timestamp: Series[datetime] = Field(alias="Time Stamp", coerce=True)
    timezone: Series[Annotated[Category, Timezone, False]] = Field(
        coerce=True, alias="Time Zone"
    )
    fuel: Series[Annotated[Category, Fuel, False]] = Field(
        alias="Fuel Category", coerce=True
    )
    gen_mw: Series[float] = Field(coerce=True, alias="Gen MW", ge=0)


# class Summary(SchemaModel):
#     date: Series[date] = Field(coerce=True)
#     fuel: Series[str] = Field(isin=[NAME[fuel] for fuel in Fuel], alias="Fuel Category")
#     gen_mw: Series[float] = Field(coerce=True)
